from sempy_labs._helper_functions import (
    resolve_item_id,
    resolve_item_name_and_id,
    resolve_workspace_id,
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    resolve_workspace_name_and_id,
)
import pandas as pd
from typing import Optional
from uuid import UUID
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def list_restore_points(
    warehouse: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Returns all restore points for a warehouse.

    This is a wrapper function for the following API: `Restore Points - List Restore Points <https://learn.microsoft.com/rest/api/fabric/warehouse/restore-points/list-restore-points>`_.

    Parameters
    ----------
    warehouse : str | uuid.UUID
        The Fabric warehouse name or ID.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing all restore points for a warehouse.
    """

    workspace_id = resolve_workspace_id(workspace)
    warehouse_id = resolve_item_id(
        item=warehouse, type="warehouse", workspace=workspace_id
    )

    columns = {
        "Restore Point Id": "string",
        "Restore Point Name": "string",
        "Description": "string",
        "Creation Mode": "string",
        "Event DateTime": "datetime",
        "Event Initiator Id": "string",
        "Event Initiator Display Name": "string",
        "Event Initiator Type": "string",
        "Event Initiator User Principal Name": "string",
    }
    df = _create_dataframe(columns=columns)
    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/warehouses/{warehouse_id}/restorePoints",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("values", []):
            cd = v.get("creationDetails", {})
            event_initiator = cd.get("eventInitiator", {})
            rows.append(
                {
                    "Restore Point Id": v.get("id"),
                    "Restore Point Name": v.get("displayName"),
                    "Description": v.get("description"),
                    "Creation Mode": v.get("creationMode"),
                    "Event DateTime": cd.get("eventDateTime"),
                    "Event Initiator Id": event_initiator.get("id"),
                    "Event Initiator Display Name": event_initiator.get("displayName"),
                    "Event Initiator Type": event_initiator.get("type"),
                    "Event Initiator User Principal Name": event_initiator.get(
                        "userDetails", {}
                    ).get("userPrincipalName"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(df=df, columns=columns)

    return df


def _resolve_restore_point_id(
    warehouse_id: UUID, restore_point: str | UUID, workspace_id: UUID
) -> str:

    df = list_restore_points(warehouse=warehouse_id, workspace=workspace_id)
    df_filt = df[
        (df["Restore Point Id"] == restore_point)
        | (df["Restore Point Name"] == restore_point)
    ]

    if df_filt.empty:
        raise ValueError(
            f"Restore Point '{restore_point}' not found in the '{warehouse_id}' warehouse within the '{workspace_id}' workspace."
        )
    return df_filt["Restore Point Id"].iloc[0]


@log
def delete_restore_point(
    warehouse: str | UUID,
    restore_point: str | UUID,
    workspace: Optional[str | UUID] = None,
):
    """
    Deletes a restore point from a warehouse.

    This is a wrapper function for the following API: `Restore Points - Delete Restore Point <https://learn.microsoft.com/rest/api/fabric/warehouse/restore-points/delete-restore-point>`_.

    Parameters
    ----------
    warehouse : str | uuid.UUID
        The Fabric warehouse name or ID.
    restore_point : str | uuid.UUID
        The restore point name or ID.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (warehouse_name, warehouse_id) = resolve_item_name_and_id(
        item=warehouse, type="Warehouse", workspace=workspace_id
    )
    restore_point_id = _resolve_restore_point_id(
        warehouse_id=warehouse_id,
        restore_point=restore_point,
        workspace_id=workspace_id,
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/warehouses/{warehouse_id}/restorePoints/{restore_point_id}",
        method="delete",
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The restore point '{restore_point}' has been deleted from the '{warehouse_name}' warehouse within the '{workspace_name}' workspace."
    )


@log
def create_restore_point(
    warehouse: str | UUID,
    name: str,
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates a restore point in a warehouse.

    This is a wrapper function for the following API: `Restore Points - Create Restore Point <https://learn.microsoft.com/rest/api/fabric/warehouse/restore-points/create-restore-point>`_.

    Parameters
    ----------
    warehouse : str | uuid.UUID
        The Fabric warehouse name or ID.
    name : str
        The name of the restore point.
    description : str, default=None
        The description of the restore point.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (warehouse_name, warehouse_id) = resolve_item_name_and_id(
        item=warehouse, type="Warehouse", workspace=workspace_id
    )

    payload = {"displayName": name}

    if description:
        payload["description"] = description

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/warehouses/{warehouse_id}/restorePoints",
        method="post",
        client="fabric_sp",
        payload=payload,
        lro_return_status_code=True,
        status_codes=[201, 202],
    )

    print(
        f"{icons.green_dot} The restore point '{name}' has been created in the '{warehouse_name}' warehouse within the '{workspace_name}' workspace."
    )


@log
def update_restore_point(
    warehouse: str | UUID,
    restore_point: str | UUID,
    name: Optional[str] = None,
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates a restore point in a warehouse.

    This is a wrapper function for the following API: `Restore Points - Update Restore Point <https://learn.microsoft.com/rest/api/fabric/warehouse/restore-points/update-restore-point>`_.

    Parameters
    ----------
    warehouse : str | uuid.UUID
        The Fabric warehouse name or ID.
    restore_point : str | uuid.UUID
        The restore point name or ID.
    name : str, default=None
        The new name of the restore point.
    description : str, default=None
        The new description of the restore point.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (warehouse_name, warehouse_id) = resolve_item_name_and_id(
        item=warehouse, type="Warehouse", workspace=workspace_id
    )
    restore_point_id = _resolve_restore_point_id(
        warehouse_id=warehouse_id,
        restore_point=restore_point,
        workspace_id=workspace_id,
    )

    payload = {}
    if name:
        payload["displayName"] = name
    if description:
        payload["description"] = description

    if not payload:
        print(
            f"{icons.yellow_dot} No updates provided for the restore point '{restore_point}'."
        )
        return

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/warehouses/{warehouse_id}/restorePoints/{restore_point_id}",
        method="patch",
        client="fabric_sp",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The restore point '{restore_point}' has been updated in the '{warehouse_name}' warehouse within the '{workspace_name}' workspace."
    )


@log
def restore_to_restore_point(
    warehouse: str | UUID,
    restore_point: str | UUID,
    workspace: Optional[str | UUID] = None,
):
    """
    Restores a warehouse in-place to the restore point specified.

    This is a wrapper function for the following API: `Restore Points - Restore To Restore Point <https://learn.microsoft.com/rest/api/fabric/warehouse/restore-points/restore-to-restore-point>`_.

    Parameters
    ----------
    warehouse : str | uuid.UUID
        The Fabric warehouse name or ID.
    restore_point : str | uuid.UUID
        The restore point name or ID.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (warehouse_name, warehouse_id) = resolve_item_name_and_id(
        item=warehouse, type="Warehouse", workspace=workspace_id
    )
    restore_point_id = _resolve_restore_point_id(
        warehouse_id=warehouse_id,
        restore_point=restore_point,
        workspace_id=workspace_id,
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/warehouses/{warehouse_id}/restorePoints/{restore_point_id}/restore",
        method="post",
        client="fabric_sp",
        lro_return_status_code=True,
        status_codes=[200, 202],
    )

    print(
        f"{icons.green_dot} The warehouse '{warehouse_name}' is being restored to the restore point '{restore_point}' within the '{workspace_name}' workspace."
    )



import pandas as pd
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    _base_api,
)


def _list_endorsements_or_favorites(type: str = "endorsements"):

    endorsement_map = {
        0: "None",
        1: "Promoted",
        2: "Certified",
        3: "Master data",
    }

    base_payload = {
        "filter": "Favorites" if "favorite" in type.lower() else "endorsement",
        "supportedTypes": [],
        "orderBy": "Default",
    }

    configs = [
        {"configurations": ["Environment", "Variables"]},
        {
            "solutions": [
                "HealthDataManager",
                "HLSCohort",
                "RetailDataManager",
                "SustainabilityDataManager",
            ]
        },
        {
            "processes": [
                "DataFlow",
                "CopyJob",
                "DatabricksCatalog",
                "DataFactory",
                "DataflowFabric",
                "DBTItem",
                "EventStream",
                "FunctionSet",
                "GraphQL",
                "KustoQueryWorkbench",
                "LLMPlugin",
                "MLExperiment",
                "MLModel",
                "MountedDataFactory",
                "MountedRelationalDatabase",
                "OperationalAgents",
                "Pipeline",
                "ApacheAirflowProject",
                "ReflexProject",
                "SparkJobDefinition",
                "SynapseNotebook",
            ]
        },
        {"insights": ["PowerBIReport", "PaginatedReport", "Dashboard"]},
        {
            "insights2": [
                "App",
                "DataExploration",
                "DigitalTwinBuilder",
                "GraphQuerySet",
                "KustoDashboard",
                "LLMPlugin",
                "Map",
                "Ontology",
                "OperationalAgents",
                "OrgApp",
            ]
        },
        {
            "data": [
                "Model",
                "Sql",
                "CosmosDB",
                "DatabricksCatalog",
                "GraphIndex",
                "KustoDatabase",
                "KustoEventHouse",
                "Lakehouse",
                "MountedRelationalDatabase",
                "PgSQLDbNative",
                "SnowflakeDatabase",
                "SqlAnalyticsEndpoint",
                "SQLDbNative",
                "Warehouse",
                "WarehouseSnapshot",
                "Lakewarehouse",
                "MountedWarehouse",
                "Datawarehouse",
            ]
        },
    ]

    rows = []
    for c in configs:
        config_name, supported_types = next(iter(c.items()))

        payload = {**base_payload, "supportedTypes": supported_types}

        response_list = _base_api(
            request="metadata/datahub/V2/artifacts",
            client="internal",
            method="post",
            payload=payload,
        ).json()

        for item in response_list:
            owner = item.get("ownerUser", {})
            artifact_endorsement = item.get("artifactEndorsement", {}).get("stage", 0)
            rows.append(
                {
                    "Item Name": item.get("displayName"),
                    "Item Id": item.get("artifactObjectId"),
                    "Workspace Name": item.get("workspaceName"),
                    "Workspace Id": item.get("workspaceObjectId"),
                    "Item Type": item.get("artifactType"),
                    "Owner": (
                        f"{owner.get('givenName')} {owner.get('familyName')}"
                        if owner
                        else None
                    ),
                    "Owner Email Address": owner.get("emailAddress"),
                    "Endorsement": endorsement_map.get(artifact_endorsement),
                    "Sensitivity Label": item.get(
                        "artifactInformationProtection", {}
                    ).get("name"),
                }
            )

    return pd.DataFrame(rows)


@log
def list_endorsements() -> pd.DataFrame:
    """
    Lists all endorsed items.

    Returns
    -------
    pandas.DataFrame
        A pandas DataFrame containing the list of endorsed items.
    """

    return _list_endorsements_or_favorites(type="endorsements")


@log
def list_favorites() -> pd.DataFrame:
    """
    Lists all favorite items.

    Returns
    -------
    pandas.DataFrame
        A pandas DataFrame containing the list of favorite items.
    """

    return _list_endorsements_or_favorites(type="favorites")



from sempy_labs._helper_functions import (
    resolve_workspace_id,
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    resolve_item_id,
    delete_item,
    resolve_workspace_name_and_id,
)
import pandas as pd
from typing import Optional
from uuid import UUID
from sempy._utils._log import log
from datetime import datetime
import sempy_labs._icons as icons


@log
def list_warehouse_snapshots(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Lists all warehouse snapshots for a given warehouse.

    This is a wrapper function for the following API: `Items - List Warehouse Snapshots <https://learn.microsoft.com/rest/api/fabric/warehousesnapshot/items/list-warehouse-snapshots>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas DataFrame containing the list of warehouse snapshots within the workspace.
    """

    workspace_id = resolve_workspace_id(workspace)

    columns = {
        "Warehouse Snapshot Id": "string",
        "Warehouse Snapshot Name": "string",
        "Description": "string",
        "Connection String": "string",
        "Parent Warehouse Id": "string",
        "Snapshot DateTime": "datetime",
    }

    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/warehousesnapshots",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Warehouse Snapshot Id": v.get("id"),
                    "Warehouse Snapshot Name": v.get("displayName"),
                    "Description": v.get("properties", {}).get("description"),
                    "Connection String": v.get("properties", {}).get(
                        "connectionString"
                    ),
                    "Parent Warehouse Id": v.get("properties", {}).get(
                        "parentWarehouseId"
                    ),
                    "Snapshot DateTime": v.get("properties", {}).get(
                        "snapshotDateTime"
                    ),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def delete_warehouse_snapshot(
    warehouse_snapshot: str | UUID,
    workspace: Optional[str | UUID] = None,
):
    """
    Deletes a warehouse snapshot in the Fabric workspace.

    This is a wrapper function for the following API: `Items - Delete Warehouse Snapshot <https://learn.microsoft.com/rest/api/fabric/warehousesnapshot/items/delete-warehouse-snapshot>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).
    """
    delete_item(item=warehouse_snapshot, type="WarehouseSnapshot", workspace=workspace)


@log
def create_warehouse_snapshot(
    name: str,
    parent_warehouse: str | UUID,
    parent_warehouse_workspace: Optional[str | UUID] = None,
    warehouse_snapshot_workspace: Optional[str | UUID] = None,
    description: Optional[str] = None,
    snapshot_datetime: Optional[datetime] = None,
):
    """
    Creates a Warehouse snapshot in the specified workspace.

    This is a wrapper function for the following API: `Items - Create Warehouse Snapshot <https://learn.microsoft.com/rest/api/fabric/warehousesnapshot/items/create-warehouse-snapshot>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The name of the warehouse snapshot to create.
    parent_warehouse : str | uuid.UUID
        The name or ID of the parent warehouse to snapshot.
    parent_warehouse_workspace : str | uuid.UUID, default=None
        The workspace name or ID where the parent warehouse is located.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    warehouse_snapshot_workspace : str | uuid.UUID, default=None
        The workspace name or ID where the warehouse snapshot will be created.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    description : str, default=None
        A description for the warehouse snapshot.
    snapshot_datetime : datetime, default=None
        The datetime for the snapshot. If not provided, the current datetime will be used.
        Example: "2024-10-15T13:00:00Z"
    """
    parent_warehouse_workspace_id = resolve_workspace_id(parent_warehouse_workspace)
    (warehouse_snapshot_workspace_name, warehouse_snapshot_workspace_id) = (
        resolve_workspace_name_and_id(warehouse_snapshot_workspace)
    )
    parent_warehouse_id = resolve_item_id(
        item=parent_warehouse, type="Warehouse", workspace=parent_warehouse_workspace_id
    )

    payload = {
        "displayName": name,
        "creationPayload": {
            "parentWarehouseId": parent_warehouse_id,
        },
    }
    if description:
        payload["description"] = description
    if snapshot_datetime:
        payload["creationPayload"]["snapshotDateTime"] = snapshot_datetime

    _base_api(
        request=f"/v1/workspaces/{warehouse_snapshot_workspace_id}/warehousesnapshots",
        client="fabric_sp",
        method="post",
        payload=payload,
        status_codes=[201, 202],
        lro_return_status_code=True,
    )

    print(
        f"{icons.green_dot} The warehouse snapshot '{name}' has been created in the workspace '{warehouse_snapshot_workspace_name}'."
    )


@log
def update_warehouse_snapshot(
    warehouse_snapshot: str | UUID,
    name: Optional[str] = None,
    description: Optional[str] = None,
    snapshot_datetime: Optional[datetime] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates the properties of a warehouse snapshot in the Fabric workspace.

    This is a wrapper function for the following API: `Items - Update Warehouse Snapshot <https://learn.microsoft.com/rest/api/fabric/warehousesnapshot/items/update-warehouse-snapshot>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    warehouse_snapshot : str | uuid.UUID
        The name or ID of the warehouse snapshot to update.
    name : str, optional
        The new name for the warehouse snapshot.
    description : str, optional
        The new description for the warehouse snapshot.
    snapshot_datetime : datetime, optional
        The new snapshot datetime for the warehouse snapshot.
        Example: "2024-10-15T13:00:00Z"
    workspace : str | uuid.UUID, optional
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """
    workspace_id = resolve_workspace_id(workspace)
    warehouse_snapshot_id = resolve_item_id(
        item=warehouse_snapshot, type="WarehouseSnapshot", workspace=workspace_id
    )

    payload = {}

    if name:
        payload["displayName"] = name
    if description:
        payload["description"] = description
    if snapshot_datetime:
        payload["properties"] = {"snapshotDateTime": snapshot_datetime}

    if not payload:
        print(
            f"{icons.yellow_dot} No updates provided for warehouse snapshot '{warehouse_snapshot}'."
        )
        return

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/warehousesnapshots/{warehouse_snapshot_id}",
        client="fabric_sp",
        method="patch",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The warehouse snapshot '{warehouse_snapshot}' has been updated."
    )



import sempy.fabric as fabric
import pandas as pd
import json
import os
from typing import Optional, List
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    _conv_b64,
    _decode_b64,
    _base_api,
    _mount,
)
from sempy_labs.lakehouse._lakehouse import lakehouse_attached
import sempy_labs._icons as icons
from sempy_labs._refresh_semantic_model import refresh_semantic_model
from uuid import UUID


@log
def create_blank_semantic_model(
    dataset: str,
    compatibility_level: int = 1702,
    workspace: Optional[str | UUID] = None,
    overwrite: bool = True,
):
    """
    Creates a new blank semantic model (no tables/columns etc.).

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    compatibility_level : int, default=1702
        The compatibility level of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    overwrite : bool, default=False
        If set to True, overwrites the existing semantic model in the workspace if it exists.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    dfD = fabric.list_datasets(workspace=workspace_id, mode="rest")
    dfD_filt = dfD[dfD["Dataset Name"] == dataset]

    if len(dfD_filt) > 0 and not overwrite:
        raise ValueError(
            f"{icons.warning} The '{dataset}' semantic model already exists within the '{workspace_name}' workspace. The 'overwrite' parameter is set to False so the blank new semantic model was not created."
        )

    min_compat = 1500
    if compatibility_level < min_compat:
        raise ValueError(
            f"{icons.red_dot} Compatiblity level must be at least {min_compat}."
        )

    # If the model does not exist
    if len(dfD_filt) == 0:
        tmsl = f"""
        {{
            "createOrReplace": {{
            "object": {{
                "database": '{dataset}'
            }},
            "database": {{
                "name": '{dataset}',
                "compatibilityLevel": {compatibility_level},
                "model": {{
                    "cultures": [
                        {{
                            "name": "en-US",
                            "linguisticMetadata": {{
                                "content": {{
                                    "Version": "1.0.0",
                                    "Language": "en-US"
                                }},
                                "contentType": "json"
                            }}
                        }}
                    ],
                    "collation": "Latin1_General_100_BIN2_UTF8",
                    "dataAccessOptions": {{
                        "legacyRedirects": true,
                        "returnErrorValuesAsNull": true,
                    }},
                    "defaultPowerBIDataSourceVersion": "powerBI_V3",
                    "sourceQueryCulture": "en-US",
                    }}
                }}
            }}
        }}
        """
    else:
        tmsl = f"""
        {{
            "createOrReplace": {{
            "object": {{
                "database": '{dataset}'
            }},
            "database": {{
                "name": '{dataset}',
                "compatibilityLevel": {compatibility_level},
                "model": {{
                "culture": "en-US",
                "defaultPowerBIDataSourceVersion": "powerBI_V3"
                }}
            }}
            }}
        }}
        """

    fabric.execute_tmsl(script=tmsl, workspace=workspace_id)

    return print(
        f"{icons.green_dot} The '{dataset}' semantic model was created within the '{workspace_name}' workspace."
    )


@log
def create_semantic_model_from_bim(
    dataset: str, bim_file: dict, workspace: Optional[str | UUID] = None
):
    """
    Creates a new semantic model based on a Model.bim file.

    This is a wrapper function for the following API: `Items - Create Semantic Model <https://learn.microsoft.com/rest/api/fabric/semanticmodel/items/create-semantic-model>`_.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    bim_file : dict
        The model.bim file.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    dfI = fabric.list_datasets(workspace=workspace_id, mode="rest")
    dfI_filt = dfI[(dfI["Dataset Name"] == dataset)]

    if not dfI_filt.empty:
        raise ValueError(
            f"{icons.red_dot} The '{dataset}' semantic model already exists as a semantic model in the '{workspace_name}' workspace."
        )

    defPBIDataset = {"version": "1.0", "settings": {}}
    payloadPBIDefinition = _conv_b64(defPBIDataset)
    payloadBim = _conv_b64(bim_file)

    payload = {
        "displayName": dataset,
        "definition": {
            "parts": [
                {
                    "path": "model.bim",
                    "payload": payloadBim,
                    "payloadType": "InlineBase64",
                },
                {
                    "path": "definition.pbidataset",
                    "payload": payloadPBIDefinition,
                    "payloadType": "InlineBase64",
                },
            ]
        },
    }

    _base_api(
        request=f"v1/workspaces/{workspace_id}/semanticModels",
        payload=payload,
        method="post",
        lro_return_status_code=True,
        status_codes=[201, 202],
    )

    print(
        f"{icons.green_dot} The '{dataset}' semantic model has been created within the '{workspace_name}' workspace."
    )


@log
def update_semantic_model_from_bim(
    dataset: str | UUID, bim_file: dict, workspace: Optional[str | UUID] = None
):
    """
    Updates a semantic model definition based on a Model.bim file.

    This is a wrapper function for the following API: `Items - Update Semantic Model Definition <https://learn.microsoft.com/rest/api/fabric/semanticmodel/items/update-semantic-model-definition>`_.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    bim_file : dict
        The model.bim file.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    defPBIDataset = {"version": "1.0", "settings": {}}
    payloadPBIDefinition = _conv_b64(defPBIDataset)
    payloadBim = _conv_b64(bim_file)

    payload = {
        "displayName": dataset_name,
        "definition": {
            "parts": [
                {
                    "path": "model.bim",
                    "payload": payloadBim,
                    "payloadType": "InlineBase64",
                },
                {
                    "path": "definition.pbidataset",
                    "payload": payloadPBIDefinition,
                    "payloadType": "InlineBase64",
                },
            ]
        },
    }

    _base_api(
        request=f"v1/workspaces/{workspace_id}/semanticModels/{dataset_id}/updateDefinition",
        payload=payload,
        method="post",
        lro_return_status_code=True,
        status_codes=None,
    )

    print(
        f"{icons.green_dot} The '{dataset_name}' semantic model has been updated within the '{workspace_name}' workspace."
    )


@log
def deploy_semantic_model(
    source_dataset: str,
    source_workspace: Optional[str | UUID] = None,
    target_dataset: Optional[str] = None,
    target_workspace: Optional[str | UUID] = None,
    refresh_target_dataset: bool = True,
    overwrite: bool = False,
    perspective: Optional[str] = None,
):
    """
    Deploys a semantic model based on an existing semantic model.

    Parameters
    ----------
    source_dataset : str
        Name of the semantic model to deploy.
    source_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    target_dataset: str
        Name of the new semantic model to be created.
    target_workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the new semantic model will be deployed.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    refresh_target_dataset : bool, default=True
        If set to True, this will initiate a full refresh of the target semantic model in the target workspace.
    overwrite : bool, default=False
        If set to True, overwrites the existing semantic model in the workspace if it exists.
    perspective : str, default=None
        Set this to the name of a perspective in the model and it will reduce the deployed model down to the tables/columns/measures/hierarchies within that perspective.
    """

    (source_workspace_name, source_workspace_id) = resolve_workspace_name_and_id(
        source_workspace
    )

    (target_workspace_name, target_workspace_id) = resolve_workspace_name_and_id(
        target_workspace
    )

    if target_dataset is None:
        target_dataset = source_dataset

    if (
        target_dataset == source_dataset
        and target_workspace_name == source_workspace_name
    ):
        raise ValueError(
            f"{icons.red_dot} The 'dataset' and 'new_dataset' parameters have the same value. And, the 'workspace' and 'new_dataset_workspace' "
            f"parameters have the same value. At least one of these must be different. Please update the parameters."
        )

    dfD = fabric.list_datasets(workspace=target_workspace_id, mode="rest")
    dfD_filt = dfD[dfD["Dataset Name"] == target_dataset]
    if not dfD_filt.empty and not overwrite:
        raise ValueError(
            f"{icons.warning} The '{target_dataset}' semantic model already exists within the '{target_workspace_name}' workspace. The 'overwrite' parameter is set to False so the source semantic model was not deployed to the target destination."
        )

    if perspective is not None:
        from sempy_labs.tom import connect_semantic_model

        with connect_semantic_model(
            dataset=source_dataset, workspace=source_workspace, readonly=True
        ) as tom:

            df_added = tom._reduce_model(perspective_name=perspective)
            bim = tom.get_bim()

    else:
        bim = get_semantic_model_bim(
            dataset=source_dataset, workspace=source_workspace_id
        )

    # Create the semantic model if the model does not exist
    if dfD_filt.empty:
        create_semantic_model_from_bim(
            dataset=target_dataset,
            bim_file=bim,
            workspace=target_workspace_id,
        )
    # Update the semantic model if the model exists
    else:
        update_semantic_model_from_bim(
            dataset=target_dataset, bim_file=bim, workspace=target_workspace_id
        )

    if refresh_target_dataset:
        refresh_semantic_model(dataset=target_dataset, workspace=target_workspace_id)

    if perspective is not None:
        return df_added


@log
def get_semantic_model_bim(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    save_to_file_name: Optional[str] = None,
) -> dict:
    """
    Extracts the Model.bim file for a given semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    save_to_file_name : str, default=None
        If specified, saves the Model.bim as a file in the lakehouse attached to the notebook.

    Returns
    -------
    dict
        The Model.bim file for the semantic model.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    bimJson = get_semantic_model_definition(
        dataset=dataset_id,
        workspace=workspace_id,
        format="TMSL",
        return_dataframe=False,
    )

    if save_to_file_name is not None:
        if not lakehouse_attached():
            raise ValueError(
                f"{icons.red_dot} In order to save the model.bim file, a lakehouse must be attached to the notebook. Please attach a lakehouse to this notebook."
            )

        local_path = _mount()
        save_folder = f"{local_path}/Files"
        file_ext = ".bim"
        if not save_to_file_name.endswith(file_ext):
            save_to_file_name = f"{save_to_file_name}{file_ext}"
        file_path = os.path.join(save_folder, save_to_file_name)
        with open(file_path, "w") as json_file:
            json.dump(bimJson, json_file, indent=4)
        print(
            f"{icons.green_dot} The {file_ext} file for the '{dataset_name}' semantic model has been saved to the lakehouse attached to the notebook within: 'Files/{save_to_file_name}'.\n\n"
        )

    return bimJson


@log
def get_semantic_model_definition(
    dataset: str | UUID,
    format: str = "TMSL",
    workspace: Optional[str | UUID] = None,
    return_dataframe: bool = True,
) -> pd.DataFrame | dict | List:
    """
    Extracts the semantic model definition.

    This is a wrapper function for the following API: `Items - Get Semantic Model Definition <https://learn.microsoft.com/rest/api/fabric/semanticmodel/items/get-semantic-model-definition>`_.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    format : str, default="TMSL"
        The output format. Valid options are "TMSL" or "TMDL". "TMSL" returns the .bim file whereas "TMDL" returns the collection of TMDL files. Can also enter 'bim' for the TMSL version.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    return_dataframe : bool, default=True
        If True, returns a dataframe.
        If False, returns the .bim file for TMSL format. Returns a list of the TMDL files (decoded) for TMDL format.

    Returns
    -------
    pandas.DataFrame | dict | List
        A pandas dataframe with the semantic model definition or the file or files comprising the semantic model definition.
    """

    valid_formats = ["TMSL", "TMDL"]

    format = format.upper()
    if format == "BIM":
        format = "TMSL"
    if format not in valid_formats:
        raise ValueError(
            f"{icons.red_dot} Invalid format. Valid options: {valid_formats}."
        )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    result = _base_api(
        request=f"v1/workspaces/{workspace_id}/semanticModels/{dataset_id}/getDefinition?format={format}",
        method="post",
        lro_return_json=True,
        status_codes=None,
    )

    files = result["definition"]["parts"]

    if return_dataframe:
        return pd.json_normalize(files)
    elif format == "TMSL":
        payload = next(
            (part["payload"] for part in files if part["path"] == "model.bim"), None
        )
        return json.loads(_decode_b64(payload))
    else:
        decoded_parts = [
            {"file_name": part["path"], "content": _decode_b64(part["payload"])}
            for part in files
        ]

        return decoded_parts


@log
def get_semantic_model_size(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Gets size of the semantic model in bytes.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    int
        The size of the semantic model in bytes
    """

    dict = fabric.evaluate_dax(
        dataset=dataset,
        workspace=workspace,
        dax_string="""
        EVALUATE SELECTCOLUMNS(FILTER(INFO.STORAGETABLECOLUMNS(), [COLUMN_TYPE] = "BASIC_DATA"),[DICTIONARY_SIZE])
        """,
    )

    used_size = fabric.evaluate_dax(
        dataset=dataset,
        workspace=workspace,
        dax_string="""
        EVALUATE SELECTCOLUMNS(INFO.STORAGETABLECOLUMNSEGMENTS(),[USED_SIZE])
        """,
    )
    dict_size = dict["[DICTIONARY_SIZE]"].sum()
    used_size = used_size["[USED_SIZE]"].sum()
    model_size = dict_size + used_size
    # Calculate proper bytes size by dividing by 1024 and multiplying by 1000 - per 1000
    if model_size >= 10**9:
        result = model_size / (1024**3) * 10**9
    elif model_size >= 10**6:
        result = model_size / (1024**2) * 10**6
    elif model_size >= 10**3:
        result = model_size / (1024) * 10**3
    else:
        result = model_size

    return result



from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
)
import pandas as pd
from sempy_labs.admin._basic_functions import list_workspaces
from sempy._utils._log import log


@log
def list_git_connections() -> pd.DataFrame:
    """
    Shows a list of Git connections.

    This is a wrapper function for the following API: `Workspaces - List Git Connections <https://learn.microsoft.com/rest/api/fabric/admin/workspaces/list-git-connections>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of Git connections.
    """

    columns = {
        "Workspace Id": "string",
        "Organization Name": "string",
        "Owner Name": "string",
        "Project Name": "string",
        "Git Provider Type": "string",
        "Repository Name": "string",
        "Branch Name": "string",
        "Directory Name": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1/admin/workspaces/discoverGitConnections",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            git = v.get("gitProviderDetails", {})
            rows.append(
                {
                    "Workspace Id": v.get("workspaceId"),
                    "Organization Name": git.get("organizationName"),
                    "Owner Name": git.get("ownerName"),
                    "Project Name": git.get("projectName"),
                    "Git Provider Type": git.get("gitProviderType"),
                    "Repository Name": git.get("repositoryName"),
                    "Branch Name": git.get("branchName"),
                    "Directory Name": git.get("directoryName"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        dfW = list_workspaces()
        df = pd.merge(
            df, dfW[["Id", "Name"]], left_on="Workspace Id", right_on="Id", how="left"
        )
        new_col_name = "Workspace Name"
        df = df.rename(columns={"Name": new_col_name})
        df.insert(1, new_col_name, df.pop(new_col_name))

        df = df.drop(columns=["Id"])

    return df



from ._items import (
    list_workspace_users,
    update_workspace_user,
    add_user_to_workspace,
    delete_user_from_workspace,
    assign_to_capacity,
    unassign_from_capacity,
    list_role_assignments,
    delete_workspace,
    get_network_communication_policy,
    set_network_communication_policy,
    get_git_outbound_policy,
    set_git_outbound_policy,
    get_outbound_cloud_connection_rules,
    get_outbound_gateway_rules,
    set_outbound_gateway_rules,
)

__all__ = [
    "list_workspace_users",
    "update_workspace_user",
    "add_user_to_workspace",
    "delete_user_from_workspace",
    "assign_to_capacity",
    "unassign_from_capacity",
    "list_role_assignments",
    "delete_workspace",
    "get_network_communication_policy",
    "set_network_communication_policy",
    "get_git_outbound_policy",
    "set_git_outbound_policy",
    "get_outbound_cloud_connection_rules",
    "get_outbound_gateway_rules",
    "set_outbound_gateway_rules",
]



from ._items import (
    list_notebooks,
    import_notebook_from_web,
    get_notebook_definition,
    create_notebook,
    update_notebook_definition,
    search_notebooks,
)

__all__ = [
    "list_notebooks",
    "import_notebook_from_web",
    "get_notebook_definition",
    "create_notebook",
    "update_notebook_definition",
    "search_notebooks",
]



from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    resolve_item_name_and_id,
    resolve_workspace_name_and_id,
    _is_valid_uuid,
)
import pandas as pd
from typing import Optional, List
from uuid import UUID
import sempy_labs._icons as icons
from sempy._utils._log import log


@log
def list_tags() -> pd.DataFrame:
    """
    Shows a list of all the tenant's tags.

    This is a wrapper function for the following API: `Tags - List Tags <https://learn.microsoft.com/rest/api/fabric/core/tags/list-tags>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of all the tenant's tags.
    """

    columns = {
        "Tag Name": "string",
        "Tag Id": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1/tags",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Tag Name": v.get("displayName"),
                    "Tag Id": v.get("id"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def resolve_tags(tags: str | List[str]) -> List[str]:
    """
    Resolves the tags to a list of strings.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    tags : str | List[str]
        The tags to resolve.

    Returns
    -------
    List[str]
        A list of resolved tags.
    """

    if isinstance(tags, str):
        tags = [tags]

    if all(_is_valid_uuid(tag) for tag in tags):
        return tags

    df = list_tags()

    tag_list = []
    for tag in tags:
        if _is_valid_uuid(tag):
            tag_list.append(tag)
        else:
            df_filt = df[df["Tag Name"] == tag]
            if df_filt.empty:
                raise ValueError(f"Tag '{tag}' not found in the tenant's tags.")
            tag_id = df_filt["Tag Id"].iloc[0]
            tag_list.append(tag_id)

    return tag_list


@log
def apply_tags(
    item: str | UUID,
    type: str,
    tags: str | UUID | List[str | UUID],
    workspace: Optional[str | UUID] = None,
):
    """
    Shows a list of all the tenant's tags.

    This is a wrapper function for the following API: `Tags - Apply Tags <https://learn.microsoft.com/rest/api/fabric/core/tags/apply-tags>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    item : str | uuid.UUID
        The name or ID of the item to apply tags to.
    type : str
        The type of the item to apply tags to. For example: "Lakehouse".
    tags : str | uuid.UUID | List[str | uuid.UUID]
        The name or ID of the tag(s) to apply to the item.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(item, type, workspace_id)

    if isinstance(tags, str):
        tags = [tags]

    tag_list = resolve_tags(tags)

    payload = {
        "tags": tag_list,
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/items/{item_id}/applyTags",
        client="fabric_sp",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} Tags {tags} applied to the '{item_name}' {type.lower()} within the '{workspace_name}' workspace"
    )


@log
def unapply_tags(
    item: str | UUID,
    type: str,
    tags: str | UUID | List[str | UUID],
    workspace: Optional[str | UUID] = None,
):
    """
    Shows a list of all the tenant's tags.

    This is a wrapper function for the following API: `Tags - Unapply Tags <https://learn.microsoft.com/rest/api/fabric/core/tags/unapply-tags>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    item : str | uuid.UUID
        The name or ID of the item to apply tags to.
    type : str
        The type of the item to apply tags to. For example: "Lakehouse".
    tags : str | uuid.UUID | List[str | uuid.UUID]
        The name or ID of the tag(s) to apply to the item.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(item, type, workspace_id)

    if isinstance(tags, str):
        tags = [tags]

    tag_list = resolve_tags(tags)

    payload = {
        "tags": tag_list,
    }

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/items/{item_id}/unapplyTags",
        client="fabric_sp",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} Tags {tags} applied to the '{item_name}' {type.lower()} within the '{workspace_name}' workspace"
    )



from ._items import (
    list_data_pipelines,
    create_data_pipeline,
    delete_data_pipeline,
    get_data_pipeline_definition,
)


__all__ = [
    "list_data_pipelines",
    "create_data_pipeline",
    "delete_data_pipeline",
    "get_data_pipeline_definition",
]



from typing import Optional, Literal
from uuid import UUID
import pandas as pd
from sempy._utils._log import log
import sempy_labs.sql_endpoint as sql


@log
def list_sql_endpoints(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the SQL endpoints within a workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the SQL endpoints within a workspace.
    """

    return sql.list_sql_endpoints(workspace=workspace)


@log
def refresh_sql_endpoint_metadata(
    item: str | UUID,
    type: Literal["Lakehouse", "MirroredDatabase"],
    workspace: Optional[str | UUID] = None,
    timeout_unit: Literal["Seconds", "Minutes", "Hours", "Days"] = "Minutes",
    timeout_value: int = 15,
) -> pd.DataFrame:
    """
    Refreshes the metadata of a SQL endpoint.

    This is a wrapper function for the following API: `Items - Refresh Sql Endpoint Metadata <https://learn.microsoft.com/rest/api/fabric/sqlendpoint/items/refresh-sql-endpoint-metadata>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    item : str | uuid.UUID
        The name or ID of the item (Lakehouse or MirroredDatabase).
    type : typing.Literal['Lakehouse', 'MirroredDatabase']
        The type of the item. Must be 'Lakehouse' or 'MirroredDatabase'.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    timeout_unit : typing.Literal['Seconds', 'Minutes', 'Hours', 'Days'], default='Minutes'
        The unit of time for the request duration before timing out. Additional duration types may be added over time.
    timeout_value : int, default=15
        The number of time units in the request duration.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the status of the metadata refresh operation.
    """

    return sql.refresh_sql_endpoint_metadata(
        item=item,
        type=type,
        workspace=workspace,
        timeout_unit=timeout_unit,
        timeout_value=timeout_value,
    )



from typing import Optional
from sempy_labs._helper_functions import (
    _base_api,
)
from sempy_labs.admin._items import (
    _resolve_item_id,
)
from uuid import UUID
from sempy._utils._log import log


@log
def export_dataflow(
    dataflow: str | UUID,
    workspace: Optional[str | UUID] = None,
) -> dict:
    """
    Shows a list of datasets for the organization.

    This is a wrapper function for the following API: `Admin - Dataflows ExportDataflowAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/dataflows-export-dataflow-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataflow : str | UUID, default=None
        The dataflow Name or Id.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or id.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
        Only used if given a dataflow name and not an id.

    Returns
    -------
    dict
        Exported Json file.
    """
    dataflow_id = _resolve_item_id(item=dataflow, type="dataflow", workspace=workspace)

    url = f"/v1.0/myorg/admin/dataflows/{dataflow_id}/export"

    response = _base_api(request=url, client="fabric_sp")

    return response.json()



from ._items import (
    list_dataflow_storage_accounts,
    assign_workspace_to_dataflow_storage,
    list_dataflows,
    list_upstream_dataflows,
    upgrade_dataflow,
    get_dataflow_definition,
    discover_dataflow_parameters,
)
from ._query_execution import (
    execute_query,
)

__all__ = [
    "list_dataflow_storage_accounts",
    "assign_workspace_to_dataflow_storage",
    "list_dataflows",
    "list_upstream_dataflows",
    "upgrade_dataflow",
    "get_dataflow_definition",
    "discover_dataflow_parameters",
    "execute_query",
]



import pandas as pd
import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    _create_dataframe,
    _conv_b64,
    delete_item,
    resolve_workspace_id,
)
from uuid import UUID
from sempy._utils._log import log


@log
def list_kql_dashboards(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the KQL dashboards within a workspace.

    This is a wrapper function for the following API: `Items - List KQL Dashboards <https://learn.microsoft.com/rest/api/fabric/kqldatabase/items/list-kql-dashboards>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the KQL dashboards within a workspace.
    """

    columns = {
        "KQL Dashboard Name": "string",
        "KQL Dashboard Id": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"v1/workspaces/{workspace_id}/kqlDashboards",
        uses_pagination=True,
        lient="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "KQL Dashboard Name": v.get("displayName"),
                    "KQL Dashboard Id": v.get("id"),
                    "Description": v.get("description"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=columns.keys())

    return df


@log
def create_kql_dashboard(
    name: str,
    content: Optional[str] = None,
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates a KQL dashboard.

    This is a wrapper function for the following API: `Items - Create KQL Dashboard <https://learn.microsoft.com/rest/api/fabric/kqldatabase/items/create-kql-dashboard>`_.

    Parameters
    ----------
    name: str
        Name of the KQL dashboard.
    content: Optional[dict], default=None
        The RealTimeDashboard.json content.
    description : str, default=None
        A description of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    payload = {"displayName": name}

    if description:
        payload["description"] = description

    if content:
        payload["definition"] = {
            "format": None,
            "parts": [
                {
                    "path": "RealTimeDashboard.json",
                    "payload": _conv_b64(content),
                    "payloadType": "InlineBase64",
                },
            ],
        }

    _base_api(
        request=f"v1/workspaces/{workspace_id}/kqlDashboards",
        method="post",
        payload=payload,
        status_codes=201,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The '{name}' KQL dashboard has been created within the '{workspace_name}' workspace."
    )


@log
def delete_kql_dashboard(
    kql_dashboard: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Deletes a KQL dashboard.

    This is a wrapper function for the following API: `Items - Delete KQL Dashboard <https://learn.microsoft.com/rest/api/fabric/kqldatabase/items/delete-kql-dashboard>`_.

    Parameters
    ----------
    kql_dashboard: str | uuid.UUID
        Name or ID of the KQL dashboard.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=kql_dashboard, item_type="KQLDashboard", workspace=workspace)



import pandas as pd
from typing import Optional, Tuple
from uuid import UUID
import sempy_labs._icons as icons
from sempy_labs.admin._basic_functions import (
    _resolve_workspace_name_and_id,
)
from sempy_labs.admin._capacities import (
    _resolve_capacity_name_and_id,
)
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    _build_url,
    _base_api,
    _create_dataframe,
)
from sempy._utils._log import log


@log
def _resolve_item_id(
    item: str,
    type: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
) -> UUID:
    if _is_valid_uuid(item):
        item_id = item

    else:
        workspace_id = _resolve_workspace_name_and_id(workspace)[1]
        dfI = list_items(workspace=workspace_id, type=type)
        dfI_filt = dfI[dfI["Item Name"] == item]

        if len(dfI_filt) == 0:
            raise ValueError(
                f"The '{item}' {type} does not exist within the '{workspace}' workspace or is not of type '{type}'."
            )

        item_id = dfI_filt["Item Id"].iloc[0]

    return item_id


@log
def _resolve_item_name_and_id(
    item: str,
    type: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
    **kwargs,
) -> Tuple[str, UUID]:
    if "item_name" in kwargs:
        print(
            "The 'item_name' parameter has been deprecated. Please replace this parameter with 'item' from the function going forward."
        )
        item = item_name
        del kwargs["item_name"]

    dfI = list_items(workspace=workspace, type=type, item=item)

    if len(dfI) > 1:
        raise ValueError(
            f"There are more than 1 item with the name '{item}'. Please specify the 'type' and/or 'workspace' to be more precise."
        )

    try:
        item_name = dfI["Item Name"].iloc[0]
        item_id = dfI["Item Id"].iloc[0]
    except Exception:
        raise ValueError(
            f"The '{item}' {type} does not exist within the '{workspace}' workspace or is not of type '{type}'."
        )

    return item_name, item_id


@log
def list_items(
    capacity: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    state: Optional[str] = None,
    type: Optional[str] = None,
    item: Optional[str | UUID] = None,
    **kwargs,
) -> pd.DataFrame:
    """
    Shows a list of active Fabric and Power BI items.

    This is a wrapper function for the following API: `Items - List Items <https://learn.microsoft.com/rest/api/fabric/admin/items/list-items>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or id.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or id.
        Defaults to None which looks into all the workspaces.
    state : str, default=None
        The item state.
    type : str, default=None
        The item type.
    item : str | UUID, default=None
        Item id or name to filter the list.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of active Fabric and Power BI items.
    """
    if "capacity_name" in kwargs:
        print(
            "The 'capacity_name' parameter has been deprecated. Please replace this parameter with 'capacity' from the function going forward."
        )
        capacity = kwargs["capacity_name"]
        del kwargs["capacity_name"]

    columns = {
        "Item Id": "string",
        "Item Name": "string",
        "Type": "string",
        "Description": "string",
        "State": "string",
        "Last Updated Date": "string",
        "Creator Principal Id": "string",
        "Creator Principal Display Name": "string",
        "Creator Principal Type": "string",
        "Creator User Principal Name": "string",
        "Workspace Id": "string",
        "Capacity Id": "string",
    }
    df = _create_dataframe(columns=columns)

    params = {}
    url = "/v1/admin/items"

    if capacity is not None:
        params["capacityId"] = _resolve_capacity_name_and_id(capacity)[1]

    if workspace is not None:
        params["workspaceId"] = _resolve_workspace_name_and_id(workspace)[1]

    if state is not None:
        params["state"] = state

    if type is not None:
        params["type"] = type

    url = _build_url(url, params)

    responses = _base_api(request=url, client="fabric_sp", uses_pagination=True)

    rows = []
    for r in responses:
        for v in r.get("itemEntities", []):
            rows.append(
                {
                    "Item Id": v.get("id"),
                    "Type": v.get("type"),
                    "Item Name": v.get("name"),
                    "Description": v.get("description"),
                    "State": v.get("state"),
                    "Last Updated Date": v.get("lastUpdatedDate"),
                    "Creator Principal Id": v.get("creatorPrincipal", {}).get("id"),
                    "Creator Principal Display Name": v.get("creatorPrincipal", {}).get(
                        "displayName"
                    ),
                    "Creator Principal Type": v.get("creatorPrincipal", {}).get("type"),
                    "Creator User Principal Name": v.get("creatorPrincipal", {})
                    .get("userDetails", {})
                    .get("userPrincipalName"),
                    "Workspace Id": v.get("workspaceId"),
                    "Capacity Id": v.get("capacityId"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    if item is not None:
        if _is_valid_uuid(item):
            df = df[df["Item Id"] == item]
        else:
            df = df[df["Item Name"] == item]

    return df


@log
def list_item_access_details(
    item: str | UUID,
    type: str,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Returns a list of users (including groups and service principals) and lists their workspace roles.

    This is a wrapper function for the following API: `Items - List Item Access Details <https://learn.microsoft.com/rest/api/fabric/admin/items/list-item-access-details>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    item : str
        Name or id of the Fabric item.
    type : str
        Type of Fabric item.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or id.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users (including groups and service principals) and lists their workspace roles.
    """

    (workspace_name, workspace_id) = _resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = _resolve_item_name_and_id(
        item=item, type=type, workspace=workspace_name
    )

    columns = {
        "User Id": "string",
        "User Name": "string",
        "User Type": "string",
        "User Principal Name": "string",
        "Item Name": "string",
        "Item Type": "string",
        "Item Id": "string",
        "Permissions": "string",
        "Additional Permissions": "string",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(
        request=f"/v1/admin/workspaces/{workspace_id}/items/{item_id}/users?type={type}",
        client="fabric_sp",
    )

    rows = []
    for v in response.json().get("accessDetails", []):
        rows.append(
            {
                "User Id": v.get("principal", {}).get("id"),
                "User Name": v.get("principal", {}).get("displayName"),
                "User Type": v.get("principal", {}).get("type"),
                "User Principal Name": v.get("principal", {})
                .get("userDetails", {})
                .get("userPrincipalName"),
                "Item Type": v.get("itemAccessDetails", {}).get("type"),
                "Permissions": v.get("itemAccessDetails", {}).get("permissions"),
                "Additional Permissions": v.get("itemAccessDetails", {}).get(
                    "additionalPermissions"
                ),
                "Item Name": item_name,
                "Item Id": item_id,
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    _build_url,
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    _is_valid_uuid,
)
from uuid import UUID
import sempy_labs._icons as icons
from sempy._utils._log import log


@log
def list_datasets(
    top: Optional[int] = None,
    filter: Optional[str] = None,
    skip: Optional[int] = None,
) -> pd.DataFrame:
    """
    Shows a list of datasets for the organization.

    This is a wrapper function for the following API: `Admin - Datasets GetDatasetsAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/datasets-get-datasets-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    top : int, default=None
        Returns only the first n results.
    filter : str, default=None
        Returns a subset of a results based on Odata filter query parameter condition.
    skip : int, default=None
        Skips the first n results.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of datasets for the organization.
    """

    columns = {
        "Dataset Id": "string",
        "Dataset Name": "string",
        "Web URL": "string",
        "Add Rows API Enabled": "bool",
        "Configured By": "string",
        "Is Refreshable": "bool",
        "Is Effective Identity Required": "bool",
        "Is Effective Identity Roles Required": "bool",
        "Target Storage Mode": "string",
        "Created Date": "datetime",
        "Content Provider Type": "string",
        "Create Report Embed URL": "string",
        "QnA Embed URL": "string",
        "Upstream Datasets": "list",
        "Users": "list",
        "Is In Place Sharing Enabled": "bool",
        "Workspace Id": "string",
        "Auto Sync Read Only Replicas": "bool",
        "Max Read Only Replicas": "int",
    }

    df = _create_dataframe(columns=columns)

    params = {}
    url = "/v1.0/myorg/admin/datasets"

    if top is not None:
        params["$top"] = top

    if filter is not None:
        params["$filter"] = filter

    if skip is not None:
        params["$skip"] = skip

    url = _build_url(url, params)
    response = _base_api(request=url, client="fabric_sp")

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "Dataset Id": v.get("id"),
                "Dataset Name": v.get("name"),
                "Web URL": v.get("webUrl"),
                "Add Rows API Enabled": v.get("addRowsAPIEnabled"),
                "Configured By": v.get("configuredBy"),
                "Is Refreshable": v.get("isRefreshable"),
                "Is Effective Identity Required": v.get("isEffectiveIdentityRequired"),
                "Is Effective Identity Roles Required": v.get(
                    "isEffectiveIdentityRolesRequired"
                ),
                "Target Storage Mode": v.get("targetStorageMode"),
                "Created Date": pd.to_datetime(v.get("createdDate")),
                "Content Provider Type": v.get("contentProviderType"),
                "Create Report Embed URL": v.get("createReportEmbedURL"),
                "QnA Embed URL": v.get("qnaEmbedURL"),
                "Upstream Datasets": v.get("upstreamDatasets", []),
                "Users": v.get("users", []),
                "Is In Place Sharing Enabled": v.get("isInPlaceSharingEnabled"),
                "Workspace Id": v.get("workspaceId"),
                "Auto Sync Read Only Replicas": v.get("queryScaleOutSettings", {}).get(
                    "autoSyncReadOnlyReplicas"
                ),
                "Max Read Only Replicas": v.get("queryScaleOutSettings", {}).get(
                    "maxReadOnlyReplicas"
                ),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def _resolve_dataset_id(dataset: str | UUID) -> str:
    if _is_valid_uuid(dataset):
        return dataset
    else:
        df = list_datasets()
        df_filt = df[df["Dataset Name"] == dataset]
        if df_filt.empty:
            raise ValueError(f"{icons.red_dot} The '{dataset}' dataset does not exist.")
        return df_filt["Dataset Id"].iloc[0]


@log
def list_dataset_users(dataset: str | UUID) -> pd.DataFrame:
    """
    Shows a list of users that have access to the specified dataset.

    This is a wrapper function for the following API: `Admin - Datasets GetDatasetUsersAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/datasets-get-dataset-users-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataset : str | uuid.UUID
        The name or ID of the dataset.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users that have access to the specified dataset.
    """

    dataset_id = _resolve_dataset_id(dataset)

    columns = {
        "User Name": "string",
        "Email Address": "string",
        "Dataset User Access Right": "string",
        "Identifier": "string",
        "Graph Id": "string",
        "Principal Type": "string",
    }

    df = _create_dataframe(columns=columns)

    url = f"/v1.0/myorg/admin/datasets/{dataset_id}/users"
    response = _base_api(request=url, client="fabric_sp")

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "User Name": v.get("displayName"),
                "Email Address": v.get("emailAddress"),
                "Dataset User Access Right": v.get("datasetUserAccessRight"),
                "Identifier": v.get("identifier"),
                "Graph Id": v.get("graphId"),
                "Principal Type": v.get("principalType"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df
